enum {

/*      PAS D'ERREUR                       */

  noer,

/*      ERREURS  DANS MP.S OU MP.C          */

  affer1, affer2, affer3, affer4, affer5,
  exger1, exger2, shier1, shier2,truer1,

  truer2, adder1, adder2, adder3, adder4,
  adder5, muler1, muler2, muler3,muler4,

  muler5, muler6, diver1, diver2, diver3,
  diver4, diver5, diver6, diver7,diver8,

  diver9, diver10, diver11, diver12, divzer1,
  dvmer1, dvmzer1, moder1, modzer1,reser1,

  reszer1, arier1, arier2, errpile, rtodber,
  gerper, divller1,

/*      ERREURS  DANS ALGLIN.C          */

  gtraner, gadmaer, gadsmaer, concater, extracter1,
  extracter2, extracter3,matextrer, matinv1, matinv2,

  mattype1, mattype2, mattype3, invmuler1, gausser1,
  kerer1, suppler1, suppler2, matqzer1, matqzer2,

  matqzer3, matqzer4, interer1, sqreder1, kerinter1,
  imagerer, amemer, maxranker, conjvecer1, conjvecer2,

/*        ERREUR DANS ANAL.C        */

  matcher1, recter1, vectmater1, unknowner1, caracer1,
  numvarer, assigner1, referer1, referer2, valencer1,

  trucer1, paramer1, eolner1, varer1, killer1,
  arrayer1, nparamer1, caseer, labeler, caseer2,

  expter1,

/*      ERREURS  DANS ARITH.C           */

  arither1, arither2, facter, puissmoder, invmoder,
  sfconter1, sfconter2, hiler1, hiler2, chiner,

  funder1, funder2, factorer1, classer1, classer2,
  qfer1, qfer2, orderer, generer, primer1,

  issquer1, reguler1, pnqner1, pnqner2, rhoer1,
  nucomper1, nupower1, biner1, biner2, bester1,
  bester2,

/*      ERREURS  DANS BASE.C            */

  hnfer1, hnfer2, hnfer3, factoreder1, factoreder2,
  allbaser1, galer1, galer2, redper1, galcer1,
  galbug1, galbug2, galbug3, galbug4, polreder1,
  initalgbug1, gpolcompbug1, rootsof1er, kerlenser, idealer1,
  idealer2, idealer3, idealer4, idealer5, initer1,
  initer2,

/*      ERREURS  DANS BIBLI.C           */

  laper1, laper2, convol1, convol2, convol3,
  grandoer1, csper1, precer1,discer1, sommeer1,

  proder1, polrecer1, vecer1, mater1, recer1,
  lller1, lllger1, lllger2, lllger3, linder1,

  linder2, algder1, recprimer, changer1, changer2,
  reorder1, reorder2, reorder3, intger2,sorter1,

  polinter1, polinter2, zbrenter1, zbrenter2, eulsumer1,
  level, vecer2, forer1, forer2, polymoder1,

  polymoder2, lllger4, lllger5, rploter1, rploter2,
  rploter3, rploter4, rploter5, rploter6, poster1,

  vecsorter1, vecsorter2,

/*      ERREURS  DANS BUCH.C           */

  bucher1, bucher2, bucher3, bucher4, bucher5,
  bucher6, bucher7, bucher8, bucher9, bucher10,
  bucher11, bucher12, bucher13, bucher14,

/*     ERREURS DANS ELLIPTIC.C         */

  elliper1, initeler1, initeler2, heller1, zeller1,
  apeller1, apeller2, apeller3, tater1, lserieser1,
  akeller1,

/*      ERREURS  DANS ES.C              */

  nomvar, quader, formater,

/*      ERREURS  DANS GEN1.C             */

  gadder1, gadder2, gadder4, gadder5, gadder6,
  gadder7, gadder9, gadder10, gadder11, gadder12,

  gadder13, gadder15, gmuler1, gmuler2, gmuler4,
  gmuler5, gmuler7, gmuler8, gmuler9, gmuler10,

  gmuler11, gmuler12, gmuler14, gdiver1, gdiver2,
  gdiver3, gdiver4, gdiver6, gdiver7, gdiver8,

  gdiver9, gdiver10, gdiver11, gdiver12, gdiver13,
  gdiver15, gdiver16, gdiver17, gdiver18, gdiver19,

  gaddbug1, gmulbug1,

      /*        ERREURS  DE  GEN2.C ET GEN3.C         */

  grefer1, mneter, gexpoer1, gexpoer2, gcmper,
  gtolger, gvaler2, gvaler4, gabser, gaffsger1,

  gaffsger2, gaffer1, gaffer2, gaffer3, gaffer4,
  gaffer5, gaffer6, gaffer7, gaffer8, gaffer9,

  gaffer10, gaffer11, gaffer12, gaffer13, gaffer14,
  gaffer15, gaffer16,gaffer17, gaffer18, gaffer19,

  gaffer20, gaffer21, gadder14, gpuier1, gpuier2,
  gpuier3, gpuier4, gdiventer, gdivmoder, bezer1,

  bezer2, ginvmoder, gmul2ner1, ginver, gnormaler,
  gmoder1, gmoder3, gmoder5, gsubser1, gsubser2,

  gsubser3, gsubser4, gsubser5, reciper, derer,
  inter1, inter2, flooer, rounder, cvtoier,

  cvtoper1, cvtoper2, rndtoier, ceiler, truncer,
  compoer1, tdeger, gsigner, iscomplexer1, denomer1,

  numerer1, gneger, lifter1, matler1, simplifyer1,
  karaer1, karaer2, minier1,

/*        ERREURS DANS GP.C       */

  includer1, inputer1, outputer1, outloger1,

/*        ERREURS DANS INIT.C       */

  memer, interrupter, nomer1, nomer2,

/*        ERREURS DANS PLOT.C       */

  ploter1, ploter2, ploter3, ploter4, ploter5,
  ploter6,

/*      ERREURS  DANS POLARIT.C         */

  poler1, poler2, poler3, poler4, poler5,
  poler6, poler7, poler8, poler9, poler10,

  poler15, bezoutpoler, factmoder, gcder1, gcder2,
  gcder3, gcder4, subrer1, subrer2, discsrer1,

  polgcder1, poltyper, sturmer2, factpoler1, factpoler2,
  henser1, newter1, rootper1, rootper2, rootper3,

  rootper4, polevaler1, polfnfer1, polfnfer2, nfisoer1,
  nfisoer2, nfisoer3, nfisoer4, rootofer1, rootofer2,


/*      ERREURS  DANS TRANS.C           */

  normer1, normer2, conjer1, realer1, imager1,
  sqrter1, sqrter2, sqrter3,sqrter4, sqrter5,

  sqrter6, exp1er, exper1, exper2, exper3,
  loger1, loger2, loger3, loger4,loger5,

  sc1er1, sc1er2, coser1, coser2, coser3,
  coser4, siner1, siner2, siner3,siner4,

  sicoer1, sicoer2, sicoer3, taner1, taner2,
  taner3, ataner1, ataner2, ataner3,ataner4,

  asiner1, asiner2, asiner3, asiner4, asiner5,
  acoser1, acoser2, acoser3,acoser4, acoser5,

  arger1, arger2, cher1, cher2, cher3,
  sher1, sher2, sher3, ther1, ther2,

  ther3, asher1, asher2, asher3, asher4,
  acher1, acher2, acher3, acher4,ather1,

  ather2, ather3, ather4,  gamer1, gamer2,
  gamer3, gamer4, gamder1, gamder2,gamder3,

  psier1, psier2, loger6, teicher1, paexper1,
  agmer1, transcer1, thetaer1, jbesselher1, zetaer1,

  zetaer2, zetaer3, polyloger1,

/*        BARATIN GENERAL            */

  talker,

/*        ERREUR PAS ENCORE IMPLEMENTE          */

  impl

};
